(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('@tensorflow/tfjs'), require('chalk'), require('@tensorflow/tfjs-node'), require('node-fetch')) :
  typeof define === 'function' && define.amd ? define(['exports', '@tensorflow/tfjs', 'chalk', '@tensorflow/tfjs-node', 'node-fetch'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.berkelium = {}, global.tf, global.chalk, global.tf, global.fetch));
})(this, (function (exports, tf, chalk, tfjsNode, fetch) { 'use strict';

  function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

  function _interopNamespace(e) {
    if (e && e.__esModule) return e;
    var n = Object.create(null);
    if (e) {
      Object.keys(e).forEach(function (k) {
        if (k !== 'default') {
          var d = Object.getOwnPropertyDescriptor(e, k);
          Object.defineProperty(n, k, d.get ? d : {
            enumerable: true,
            get: function () { return e[k]; }
          });
        }
      });
    }
    n["default"] = e;
    return Object.freeze(n);
  }

  var tf__namespace = /*#__PURE__*/_interopNamespace(tf);
  var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
  var fetch__default = /*#__PURE__*/_interopDefaultLegacy(fetch);

  function _defineProperty(obj, key, value) {
    key = _toPropertyKey(key);
    if (key in obj) {
      Object.defineProperty(obj, key, {
        value: value,
        enumerable: true,
        configurable: true,
        writable: true
      });
    } else {
      obj[key] = value;
    }
    return obj;
  }
  function _toPrimitive(input, hint) {
    if (typeof input !== "object" || input === null) return input;
    var prim = input[Symbol.toPrimitive];
    if (prim !== undefined) {
      var res = prim.call(input, hint || "default");
      if (typeof res !== "object") return res;
      throw new TypeError("@@toPrimitive must return a primitive value.");
    }
    return (hint === "string" ? String : Number)(input);
  }
  function _toPropertyKey(arg) {
    var key = _toPrimitive(arg, "string");
    return typeof key === "symbol" ? key : String(key);
  }

  const PUNCTUATIONS = ['.', '\'', '"', '\\', ',', '{', '}', '[', ']', ':', ';', '|', '<', '>', '?', '/', '~', '`', '!', '@', '#', '#', '$', '%', '^', '&', '*', '(', ')', '_', '-', '+', '='];

  const ENGLISH_CHAR = {
    'a': 10,
    'b': 11,
    'c': 12,
    'd': 13,
    'e': 14,
    'f': 15,
    'g': 16,
    'h': 17,
    'i': 18,
    'j': 19,
    'k': 20,
    'l': 21,
    'm': 22,
    'n': 23,
    'o': 24,
    'p': 25,
    'q': 26,
    'r': 27,
    's': 28,
    't': 29,
    'u': 30,
    'v': 31,
    'w': 32,
    'x': 33,
    'y': 34,
    'z': 35
  };

  class Tokenizer {
    constructor() {
      _defineProperty(this, "language", 'en');
      this.language = 'en';
    }
    async initialize(sentence) {
      const rawString = sentence.toLowerCase();
      let stringTokens = await this.whiteSpaceToken(rawString);
      stringTokens = await this.removePunctuations(stringTokens);
      stringTokens = await this.normalizeText(stringTokens);
      return stringTokens;
    }
    whiteSpaceToken(sentence) {
      return sentence.split(' ');
    }
    removePunctuations(tokens) {
      const token_collection = [];
      tokens.map(token => {
        if (token !== '') {
          PUNCTUATIONS.map(symbol => {
            token = token.replace(symbol, '');
            // console.log(symbol, token);
          });

          token_collection.push(token);
        }
      });
      return token_collection;
    }
    normalizeText(sentence) {
      let normalizedString = [];
      if (this.language === 'en') {
        sentence.map(async word => {
          const isPure = await this.checkLanguageChar(word);
          if (isPure) {
            normalizedString.push(word);
          }
        });
      }
      return normalizedString;
    }
    checkLanguageChar(word) {
      const isPure = new Promise((resolve, reject) => {
        try {
          for (let index = 0; index < word.length; index++) {
            const isChar = ENGLISH_CHAR.hasOwnProperty(word.charAt(index));
            const hasNumber = parseInt(word.charAt(index));
            if (!isChar && isNaN(hasNumber)) {
              // console.log(index, hasNumber, typeof hasNumber);
              resolve(false);
            }
            if (index === word.length - 1) {
              resolve(true);
            }
          }
        } catch (error) {
          reject(error);
          console.log(error);
        }
      });
      return isPure;
    }
  }

  //import { ENGLISH_CHAR } from './lang/english';

  class TextEncorder {
    //private language = 'en';

    constructor() {
      //this.language = 'en';
    }
    async encode(tokens) {
      let encodedTokens = [];
      const vocab = {
        'oov': 0
      };

      /* if (this.language === 'en') {
          tokens.map(async (word: string)=>{
              const encoded = await this.generateEncodings(word);
              encodedTokens.push(encoded);
          });
      } */

      return new Promise((resolve, reject) => {
        try {
          const uniqueVocab = tokens.map(word => {
            if (!encodedTokens.includes(word)) {
              encodedTokens.push(word);
            }
          });
          Promise.all(uniqueVocab).then(() => {
            encodedTokens.sort();
            encodedTokens.map((word, index) => {
              vocab[word] = index + 1;
            });
            // console.log(encodedTokens);
          });

          resolve(vocab);
        } catch (error) {
          reject(error);
        }
      });
    }

    /* private generateEncodings(word: string) {
        const encodedToken = new Promise((resolve, reject) => {
            try {
                const encoding = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
                let word_length = encoding.length;
                if (word.length <= word_length) {
                    word_length = word.length;
                }
                  for(let index = 0; index < word.length; index++) {
                    let charID = 0;
                    const hasNumber = parseInt(word.charAt(index));
                    // console.log(typeof hasNumber);
                    if (!isNaN(hasNumber)) {
                        charID = parseInt(word.charAt(index));
                    } else {
                        charID = ENGLISH_CHAR[word.charAt(index)];
                    }
                    // console.log(charID, word.charAt(index), typeof charID);
                    encoding[index] = charID;
                }
                resolve(encoding);               
            } catch (error) {
                reject(error);
                console.log(error);
            }
        });
          return encodedToken;
    } */
  }

  class Preprocessor {
    constructor(rawDataset) {
      _defineProperty(this, "RAW_DATA", []);
      _defineProperty(this, "CORPUS_DATA", []);
      _defineProperty(this, "INTENT_PATTERNS", []);
      _defineProperty(this, "INTENT_CLASSES", []);
      _defineProperty(this, "TRAINING_DATA", []);
      _defineProperty(this, "TRAINING_LABELS", []);
      _defineProperty(this, "INPUT_LENGTH", 10);
      _defineProperty(this, "VOCABULARY", void 0);
      this.RAW_DATA = rawDataset;
    }
    async init() {
      let trainingDataset = {
        x: this.TRAINING_DATA,
        y: this.INTENT_CLASSES,
        labels: this.TRAINING_LABELS,
        vocab: this.VOCABULARY,
        length: this.INPUT_LENGTH
      };
      try {
        await this.prepareTrainingData();
        await this.generateVocabulary(this.CORPUS_DATA.flat());
        await this.textVectorize();
        await this.padSequence();
        trainingDataset = {
          x: this.TRAINING_DATA,
          y: this.INTENT_CLASSES,
          labels: this.TRAINING_LABELS,
          vocab: this.VOCABULARY,
          length: this.INPUT_LENGTH
        };
      } catch (error) {
        console.log(chalk__default["default"].red(' error ') + chalk__default["default"].redBright(' Data read error: '), error);
      }
      return trainingDataset;
    }
    async prepareTrainingData() {
      console.log(chalk__default["default"].bgGreen.black(' info ') + chalk__default["default"].greenBright(` Data Preprocessing Started: ${this.RAW_DATA.length} inputs\n`));
      try {
        const intentLabels = this.RAW_DATA.map(dataObj => {
          if (!this.TRAINING_LABELS.includes(dataObj[1])) {
            this.TRAINING_LABELS.push(dataObj[1]);
          }
          this.CORPUS_DATA.push(dataObj[0]);
        });
        Promise.all(intentLabels).then(async () => {
          return;
        });
      } catch (error) {
        console.error(chalk__default["default"].redBright('Encoding Error:'), error);
      }
    }
    generateVocabulary(textData) {
      const bkTokenizer = new Tokenizer();
      const tokenCollection = [];
      return new Promise(resolve => {
        const getVocab = textData.map(async sentence => {
          let tokens = await bkTokenizer.initialize(sentence);
          tokenCollection.push(tokens);
        });
        Promise.all(getVocab).then(async () => {
          const bkEncoder = new TextEncorder();
          this.VOCABULARY = await bkEncoder.encode(tokenCollection.flat());
          // console.log(Object.keys(this.VOCABULARY).length);
          resolve(this.VOCABULARY);
        });
      });
    }
    textVectorize() {
      const bkTokenizer = new Tokenizer();
      return new Promise((resolve, reject) => {
        const encodedDS = this.RAW_DATA.map(async dataObj => {
          const patternEmbeddings = await bkTokenizer.initialize(dataObj[0]);
          const classEmbeddings = await this.classOneHotEncode(dataObj[1]);
          if (this.INPUT_LENGTH < patternEmbeddings.length) {
            this.INPUT_LENGTH = patternEmbeddings.length;
            //console.log(patternEmbeddings.length);
          }

          this.INTENT_PATTERNS.push(patternEmbeddings);
          this.INTENT_CLASSES.push(classEmbeddings);
          // console.log(`sample encoding?: ${patternEmbeddings.length}`);
        });

        Promise.all(encodedDS).then(() => {
          resolve(this.INTENT_PATTERNS);
        }).catch(err => {
          console.log('Data Error:', err);
          reject(err);
        });
      });
    }
    padSequence() {
      return new Promise((resolve, reject) => {
        const paddedSequences = this.INTENT_PATTERNS.map(tokens => {
          const numPad = this.INPUT_LENGTH - tokens.length;
          const zeroPads = tf__namespace.zeros([1, numPad]).arraySync();
          // console.log(`pad?: ${numPad} => ${tokens.length} ${typeof zeroPads}`);
          try {
            let finalEmbedding = [];
            const mapToken = tokens.map(word => {
              finalEmbedding.push(this.VOCABULARY[word]);
              // console.log(word);
            });

            Promise.all(mapToken).then(() => {
              if (numPad !== 0) {
                this.TRAINING_DATA.push(finalEmbedding.concat(zeroPads[0]));
              } else {
                this.TRAINING_DATA.push(finalEmbedding);
              }
              // console.log(this.TRAINING_DATA);
            }).catch(err => {
              console.log('Padding Error:', err);
              reject(err);
            });
          } catch (error) {
            console.log('Vectorizing Error:', error);
          }
        });
        Promise.all(paddedSequences).then(() => {
          // console.log(this.TRAINING_DATA);
          resolve(this.TRAINING_DATA);
        }).catch(err => {
          console.log('Padding Error:', err);
          reject(err);
        });
      });
    }
    classOneHotEncode(className) {
      let classVector = tf__namespace.zeros([1, this.TRAINING_LABELS.length]).arraySync();
      const classIndex = this.TRAINING_LABELS.indexOf(className);
      classVector[0][classIndex] = 1;
      // console.log(classVector);
      return classVector[0];
    }
  }

  class IntentClassification {
    constructor(dataset) {
      _defineProperty(this, "MODEL", void 0);
      _defineProperty(this, "INTENT_PATTERNS", []);
      _defineProperty(this, "INTENT_CLASSES", []);
      _defineProperty(this, "TRAINING_LABELS", []);
      _defineProperty(this, "TRAIN_EPOCHS", 20);
      _defineProperty(this, "INPUT_LENGTH", 10);
      _defineProperty(this, "VOCABULARY", void 0);
      this.INTENT_PATTERNS = dataset.x;
      this.INTENT_CLASSES = dataset.y;
      this.TRAINING_LABELS = dataset.labels;
      this.INPUT_LENGTH = dataset.length;
      this.VOCABULARY = dataset.vocab;
    }
    train() {
      // console.log(this.INPUT_LENGTH);
      return new Promise(async (resolve, reject) => {
        const trainX = tf__namespace.stack(this.INTENT_PATTERNS.map(x => tf__namespace.tensor1d(x)));
        const trainY = tf__namespace.stack(this.INTENT_CLASSES.map(x => tf__namespace.tensor1d(x)));
        const length = this.TRAINING_LABELS.length;
        // const batchSize = this.INTENT_PATTERNS.length;
        // const xyDataset = tf.data.zip({ xs: trainX, ys: trainY }).batch(batchSize);
        const inputShape = [this.INPUT_LENGTH];

        // console.log(`dataset: ${trainX.shape}`);

        const embeddings = tf__namespace.layers.embedding({
          inputDim: Object.keys(this.VOCABULARY).length,
          outputDim: 16,
          name: 'bkEmbed',
          inputShape: inputShape,
          maskZero: true
        });
        this.MODEL = tf__namespace.sequential();
        this.MODEL.add(embeddings);
        // this.MODEL.add(tf.layers.globalAveragePooling1d());
        // this.MODEL.add(tf.layers.bidirectional({ layer: tf.layers.lstm({ units: 64 }) }));
        this.MODEL.add(tf__namespace.layers.flatten());
        this.MODEL.add(tf__namespace.layers.dense({
          units: 64,
          activation: 'relu'
        }));
        // this.MODEL.add(tf.layers.dropout({ rate: 0.5 }));
        this.MODEL.add(tf__namespace.layers.dense({
          units: length,
          activation: 'sigmoid'
        }));
        this.MODEL.compile({
          optimizer: 'adam',
          loss: 'categoricalCrossentropy',
          metrics: ['accuracy']
        });
        // console.log(trainX.shape, this.INTENT_PATTERNS[5], this.INTENT_PATTERNS[5].length);

        this.MODEL.summary();
        try {
          await this.MODEL.fit(trainX, trainY, {
            epochs: this.TRAIN_EPOCHS,
            verbose: 1,
            callbacks: tf__namespace.callbacks.earlyStopping({
              monitor: 'acc'
            })
          }).then(info => {
            const infoIndex = info.epoch.length - 1;
            const finalLoss = info.history.loss[infoIndex].toFixed(5);
            const finalAcc = info.history.acc[infoIndex].toFixed(5);
            console.log(chalk__default["default"].bgYellow.black(`\n Berkelium `) + chalk__default["default"].greenBright(` Training Completed at ${info.epoch.length} ==>`) + chalk__default["default"].yellowBright(` Loss: ${finalLoss}`) + chalk__default["default"].green(` | Accuracy: ${finalAcc}\n`));
            resolve(this.MODEL);
          });
        } catch (err) {
          console.log('Training Err:', err);
          reject(err);
        }
      });
    }
  }

  class IntentPrediction {
    constructor(modelURL) {
      _defineProperty(this, "MODEL_URL", void 0);
      _defineProperty(this, "THE_MODEL", void 0);
      _defineProperty(this, "THE_METADATA", void 0);
      this.MODEL_URL = modelURL;
    }
    loadModel() {
      return new Promise(async (resolve, reject) => {
        try {
          this.THE_MODEL = await tf__namespace.loadLayersModel(`${this.MODEL_URL}/model.json`);
          fetch__default["default"](`${this.MODEL_URL}/metadata.json`).then(response => response.json()).then(data => {
            this.THE_METADATA = data;
            resolve(this.THE_MODEL);
          });
        } catch (error) {
          console.log(chalk__default["default"].red(`Oops! We couldn't load your model:\n${error}`));
          reject(error);
        }
      });
    }
    predict(sentence) {
      return new Promise(async (resolve, reject) => {
        try {
          const inputVector = await this.input(sentence);
          const reply = await this.THE_MODEL.predict(tf__namespace.tensor([inputVector]));
          const prediction = reply.arraySync();
          let max = Math.max(...prediction[0]);
          let index = prediction[0].indexOf(max);
          // console.log(prediction[0], index, max, prediction[0].length);
          const predictionData = {
            max: max,
            index: index,
            predictions: reply,
            intent: this.THE_METADATA.intents[index]
          };
          resolve(predictionData);
        } catch (error) {
          console.log(chalk__default["default"].red(`Oops! Prediction error:\n${error}`));
          reject(error);
        }
      });
    }
    async input(sentence) {
      const bkTokenizer = new Tokenizer();
      const tokens = await bkTokenizer.initialize(sentence);
      const textVector = tokens.map(word => {
        let vector;
        if (this.THE_METADATA.vocab.hasOwnProperty(word)) {
          vector = this.THE_METADATA.vocab[word];
        } else {
          vector = 0;
        }
        return vector;
      });
      return new Promise((resolve, reject) => {
        Promise.all(textVector).then(() => {
          const inputVector = this.padSequence(textVector);
          resolve(inputVector);
        }).catch(err => {
          console.log(chalk__default["default"].red(`input sentence error:\n${err}`));
          reject(err);
        });
      });
    }
    padSequence(textVector) {
      let finalencode;
      if (textVector.length < this.THE_METADATA.seq_len) {
        const numPad = this.THE_METADATA.seq_len - textVector.length;
        const padzero = tf__namespace.zeros([1, numPad]).arraySync();
        finalencode = textVector.concat(padzero[0]);
      } else {
        finalencode = textVector.slice(0, this.THE_METADATA.seq_len - 1);
      }
      return finalencode;
    }
  }

  let newModel;
  const berkelium = {
    /**
     * Tokenizer function is used to tokenize
     * text input and returns a `Array<string>` of tokens.
     * @param sentence `string` A string input to be tokenized.
     * @returns `Array<string>` Returns a string array of tokens.
     */
    tokenize: sentence => {
      const bkTokenizer = new Tokenizer();
      return bkTokenizer.initialize(sentence);
    },
    /**
     * Encoder function vectorize the string tokens. Which means,
     * each string tokens will be assigned a unique number.
     * This is important when preparing text data to be feed in to a machine learning model.
     * Usually this function is used after tokenizing your string data using `tokenize` function.
     * @param tokens `Array<string>` A string array of tokens.
     * @returns `DICTIONARY_BOOK` An dictionary of vocabulary with their assigned unique numeric value.
     */
    encode: tokens => {
      const encodings = new TextEncorder();
      return encodings.encode(tokens);
    },
    /**
     * Preprocessor function process the text data and create a `DATASET` object 
     * that can be used to train machine learning model.
     * @param rawData `Array<Array<string>>` training data
     * @returns `DATASET` returns a dataset object
     */
    preprocess: async rawData => {
      const preProcessor = new Preprocessor(rawData);
      const preProcessedData = await preProcessor.init();
      return preProcessedData;
    },
    multiClass: {
      /**
       * Use to train a model
       * @param dataset accept `DATASET` Object.
       * @returns `MODEL` Object
       */
      train: async dataset => {
        try {
          const bkTrain = new IntentClassification(dataset);
          const model = await bkTrain.train();
          return model;
        } catch (error) {
          console.log(chalk__default["default"].red(' error ') + chalk__default["default"].redBright(' Training failed: '), error);
        }
      },
      /**
       * Use to load a trained model
       * Note: `loadModel` function only accepts absolute URLs.
       * @param modelURL Absolute URL to model location
       * @returns `MODEL` Object
       */
      loadModel: async modelURL => {
        newModel = new IntentPrediction(modelURL);
        await newModel.loadModel();
        return newModel;
      }
    }
  };

  exports.berkelium = berkelium;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
