import { DATASET } from './core/preprocess';
import { IntentPrediction } from './modules/intent-prediction';
export declare const berkelium: {
    /**
     * Tokenizer function is used to tokenize
     * text input and returns a `Array<string>` of tokens.
     * @param sentence `string` A string input to be tokenized.
     * @returns `Array<string>` Returns a string array of tokens.
     */
    tokenize: (sentence: string) => Promise<string[]>;
    /**
     * Encoder function vectorize the string tokens. Which means,
     * each string tokens will be assigned a unique number.
     * This is important when preparing text data to be feed in to a machine learning model.
     * Usually this function is used after tokenizing your string data using `tokenize` function.
     * @param tokens `Array<string>` A string array of tokens.
     * @returns `DICTIONARY_BOOK` An dictionary of vocabulary with their assigned unique numeric value.
     */
    encode: (tokens: Array<string>) => Promise<unknown>;
    /**
     * Preprocessor function process the text data and create a `DATASET` object
     * that can be used to train machine learning model.
     * @param rawData `Array<Array<string>>` training data
     * @returns `DATASET` returns a dataset object
     */
    preprocess: (rawData: Array<any>) => Promise<DATASET>;
    multiClass: {
        /**
         * Use to train a model
         * @param dataset accept `DATASET` Object.
         * @returns `MODEL` Object
         */
        train: (dataset: DATASET) => Promise<unknown>;
        /**
         * Use to load a trained model
         * Note: `loadModel` function only accepts absolute URLs.
         * @param modelURL Absolute URL to model location
         * @returns `MODEL` Object
         */
        loadModel: (modelURL: string) => Promise<IntentPrediction>;
    };
};
